using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Xml.Serialization;
using log4net;

namespace Orbiter.Multiplayer
{
    /// <summary>
    /// The main server class.
    /// </summary>
    public class Server
    {
        private string configurationFileName;
        private OMPServerConfiguration configuration;
        private NTPClient ntpClient;
        private Listener listener;
        private Web web;
        private Map map;
        private Time time;
        private bool locked;
        private static readonly XmlSerializer ConfigurationSerializer = new XmlSerializer(typeof(OMPServerConfiguration));
        private FileSystemWatcher configurationFileWatcher = new FileSystemWatcher();
        private static readonly ILog log = LogManager.GetLogger("    ");
        private NTPTime lastConfigChange;
        private int configChangeAttempt;

        public OMPServerConfiguration Configuration
        {
            get { return configuration; }
        }

        public Map Map
        {
            get { return map; }
        }
        
        public NTPClient NTP
        {
            get { return ntpClient; }
        }

        public Server(string[] args)
        {
            lastConfigChange=new NTPTime();
            configChangeAttempt = 0;
            configurationFileName = "server.xml";
            if (args.Length > 0) configurationFileName = args[0];
        }

        public Listener Listener
        {
            get { return listener; }
        }

        public void Start()
        {
            //Read configuration file
            if (!Path.IsPathRooted(configurationFileName)) configurationFileName = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, configurationFileName);
            if (!File.Exists(configurationFileName))
                throw new FileNotFoundException("Configuration file '" + configurationFileName + "' not found!");
            using (var file=File.OpenRead(configurationFileName))
                configuration = ConfigurationSerializer.Deserialize(file) as OMPServerConfiguration;

            //Generate time
            time=new Time(this);
            
            //Start NTP system
            ntpClient = new NTPClient(  configuration.NTP.SamplesSpecified ? configuration.NTP.Samples : 30,
                                        configuration.NTP.HistorySpecified ? configuration.NTP.History : 60,
                                        configuration.Timing.SNTPSpecified ? configuration.Timing.SNTP : 250,
                                        configuration.Timing.ResyncSpecified ? configuration.Timing.Resync : 2000,
                                        configuration.NTP.OffsetSpecified ? configuration.NTP.Offset : 0,
                                        configuration.NTP.SkewSpecified ? configuration.NTP.Skew / 1E6 : 0,
                                        10);
            if (configuration.NTP.Items!=null)
                foreach (var server in configuration.NTP.Items)
                {
                    ntpClient.AddServer(new NTPServerInfo(server.Address)
                    {
                        Delay = server.DelaySpecified ? server.Delay : 3,
                        Misses = server.MissesSpecified ? server.Misses : 3,
                        MissesAlarm = server.AlarmSpecified ? server.Alarm : 3
                    });
                }
            ntpClient.Start();

            //Generate map
            map = new Map(this);
            map.Reconfigure();

            //Start TCP listener
            listener =new Listener(this);
            listener.Reconfigure();

            //Start web server
            web = new Web(this);
            web.Reconfigure();

            var configFile = Path.GetFullPath(configurationFileName);
            configurationFileWatcher.Path = Path.GetDirectoryName(configFile);
            configurationFileWatcher.Filter = Path.GetFileName(configFile);
            configurationFileWatcher.NotifyFilter = NotifyFilters.LastWrite;
            configurationFileWatcher.Changed += ConfigurationFile_Changed;
            configurationFileWatcher.EnableRaisingEvents = true;
        }

        void ConfigurationFile_Changed(object sender, FileSystemEventArgs e)
        {
            if (e.ChangeType != WatcherChangeTypes.Changed) return;
            var lastChangeAge = ntpClient.Now - lastConfigChange;
            lastConfigChange = ntpClient.Now;
            configChangeAttempt = lastChangeAge > 5 ? 0 : configChangeAttempt + 1;
            var prefix = configChangeAttempt > 0 ? string.Format("Retry {0}: ", configChangeAttempt) : "";
            OMPServerConfiguration newConfiguration;
            try
            {
                using (var file = File.OpenRead(configurationFileName))
                    try
                    {
                        newConfiguration = ConfigurationSerializer.Deserialize(file) as OMPServerConfiguration;
                    }
                    catch (InvalidOperationException ex)
                    {
                        log.Error(prefix + "Deserializing changed configuration failed!", ex);
                        return;
                    }
            }
            catch (IOException)
            {
                log.Info(prefix + "Unable to read changed configuration.");
                return;
            }
            log.Info(prefix + "Reading changed configuration...");
            configuration = newConfiguration;
            Reconfigure();
        }
        
        public void Stop()
        {
            configurationFileWatcher.EnableRaisingEvents = false;
            configurationFileWatcher.Changed -= ConfigurationFile_Changed;
            web.Stop();
            listener.Stop();
            ntpClient.Stop();
            map.Clear();
            var skew=(int)(ntpClient.Skew * 1E6);
            if (configuration.NTP.Offset == ntpClient.Offset && configuration.NTP.Skew == skew) return;
            configuration.NTP.OffsetSpecified = (configuration.NTP.Offset = ntpClient.Offset) != 0;
            configuration.NTP.SkewSpecified = (configuration.NTP.Skew = skew) != 0;
            using (var file=new FileStream(configurationFileName, FileMode.Create))
                ConfigurationSerializer.Serialize(file, configuration);
        }

        public void Reconfigure()
        {
            //Reconfigure NTP client
            ntpClient.SampleWaitTime = configuration.Timing.SNTPSpecified ? configuration.Timing.SNTP : 250;
            ntpClient.CycleWaitTime = configuration.Timing.ResyncSpecified ? configuration.Timing.Resync : 2000;
            var toBeRemoved = new List<NTPServerInfo>(ntpClient.Servers);
            if (configuration.NTP.Items!=null)
                foreach (var s in configuration.NTP.Items)
                {
                    var found = false;
                    foreach(var server in ntpClient.Servers)
                        if (s.Address==server.Address)
                        {
                            toBeRemoved.Remove(server);
                            server.Delay = s.DelaySpecified ? s.Delay : 3;
                            server.Misses = s.MissesSpecified ? s.Misses : 3;
                            server.MissesAlarm = s.AlarmSpecified ? s.Alarm : 3;
                            found = true;
                            break;
                        }
                    if (!found)
                        ntpClient.AddServer(new NTPServerInfo(s.Address)
                        {
                            Delay = s.DelaySpecified ? s.Delay : 3,
                            Misses = s.MissesSpecified ? s.Misses : 3,
                            MissesAlarm = s.AlarmSpecified ? s.Alarm : 3
                        });
                }
            foreach (var server in toBeRemoved) ntpClient.RemoveServer(server);

            map.Reconfigure();
            listener.Reconfigure();
            web.Reconfigure();
        }

        public Time Time
        {
            get { return time; }
        }

        public bool Locked
        {
            get { return locked; }
            set { locked = value; }
        }
    }
}