using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using log4net;

namespace Orbiter.Multiplayer
{
    /// <summary>
    /// SNTP client for retrieving time information from NTP servers.
    /// </summary>
    public class NTPClient
    {
        private double offset;
        private NTPTime baseTime;
        private double offsetAdjustment;
        private double skew;
        private double skewMinimum;
        private double skewMaximum;
        private List<NTPServerInfo> servers=new List<NTPServerInfo>();
        private Thread synchronizerThread;
        private bool synchronizing;
        private int sntpSamples;
        private int sntpHistory;
        private int sampleWaitFirstHalfTime;
        private int sampleWaitSecondHalfTime;
        private int cycleWaitFirstTimePart;
        private int cycleWaitSecondTimePart;
        private static readonly ILog log = LogManager.GetLogger("SNTP");
        private ManualResetEvent added = new ManualResetEvent(true);

        /// <summary>
        /// Gets the offset.
        /// </summary>
        /// <value>The offset.</value>
        public double Offset
        {
            get { return offset; }
        }

        /// <summary>
        /// Gets the skew.
        /// </summary>
        /// <value>The skew.</value>
        public double Skew
        {
            get { return skew; }
        }

        /// <summary>
        /// Gets the skew minimum.
        /// </summary>
        /// <value>The skew minimum.</value>
        public double SkewMinimum
        {
            get { return skewMinimum; }
        }

        /// <summary>
        /// Gets the skew maximum.
        /// </summary>
        /// <value>The skew maximum.</value>
        public double SkewMaximum
        {
            get { return skewMaximum; }
        }

        /// <summary>
        /// Adds a server.
        /// </summary>
        /// <param name="server">The server.</param>
        public void AddServer(NTPServerInfo server)
        {
            servers.Add(server);
            added.Set();
        }

        /// <summary>
        /// Removes a server.
        /// </summary>
        /// <param name="server">The server.</param>
        public void RemoveServer(NTPServerInfo server)
        {
            servers.Remove(server);
        }

        /// <summary>
        /// Gets the servers.
        /// </summary>
        /// <value>The servers.</value>
        public List<NTPServerInfo> Servers
        {
            get { return servers; }
        }

        /// <summary>
        /// Gets the server time.
        /// </summary>
        /// <param name="hostName">Name of the host.</param>
        /// <returns></returns>
        public NTPServerResponse GetServerTime(string hostName)
        {
            return GetServerTime(hostName, 123);
        }

        /// <summary>
        /// Gets the server time.
        /// </summary>
        /// <param name="hostName">Name of the host.</param>
        /// <param name="port">The port.</param>
        /// <returns></returns>
        public NTPServerResponse GetServerTime(string hostName, int port)
        {
            lock (this)
            {
                //Connect socket
                IntPtr buffer;
                byte[] rawdatas;
                Socket pSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                pSocket.ReceiveTimeout = 2000;
                pSocket.SendTimeout = 2000;
                pSocket.Connect(new IPEndPoint(Dns.GetHostAddresses(hostName)[0], port));

                //Initialise the NTPBasicInfo packet
                NTPBasicInfo nbi = new NTPBasicInfo();
                nbi.VersionMode = 27; //Encoded NTP client request and NTP version 3.0
                if (baseTime == null) baseTime = NTPTime.Now;
                nbi.TransmitTimestamp = Now;

                //Send NTPBasicInfoPacket
                rawdatas = new byte[Marshal.SizeOf(nbi)];
                buffer = Marshal.AllocHGlobal(rawdatas.Length);
                Marshal.StructureToPtr(nbi, buffer, false);
                Marshal.Copy(buffer, rawdatas, 0, rawdatas.Length);
                Marshal.FreeHGlobal(buffer);
                pSocket.Send(rawdatas);

                //Read server response as NTPFullPacket
                rawdatas = new byte[Marshal.SizeOf(new NTPFullPacket())];
                pSocket.Receive(rawdatas);
                buffer = Marshal.AllocHGlobal(rawdatas.Length);
                Marshal.Copy(rawdatas, 0, buffer, rawdatas.Length);
                NTPFullPacket nfp = (NTPFullPacket)Marshal.PtrToStructure(buffer, typeof(NTPFullPacket));
                Marshal.FreeHGlobal(buffer);

                //Disconnect from server
                pSocket.Close();

                //Create response
                NTPTime originate = new NTPTime(nfp.Basic.OriginateTimestamp);
                NTPTime receive = new NTPTime(nfp.Basic.ReceiveTimestamp);
                NTPTime transmit = new NTPTime(nfp.Basic.TransmitTimestamp);
                NTPTime destination = Now;
                return new NTPServerResponse(
                    (nfp.Basic.VersionMode & 0xC0) >> 6,
                    nfp.Basic.Stratum,
                    new [] { originate, receive, transmit, destination },
                    ((destination - originate) - (receive - transmit)),
                    ((receive - originate) + (transmit - destination)) / 2);
            }
        }

        /// <summary>
        /// Gets the current compensated time.
        /// </summary>
        /// <value>The current compensated time.</value>
        public NTPTime Now
        {
            get
            {
                lock (this)
                {
                    var t = NTPTime.Now;
                    var d = t - baseTime;
                    var dA = Math.Max(d, CycleWaitTime);
                    t = t + offset + d*skew + dA*offsetAdjustment;
                    return t;
                }
            }
        }

        /// <summary>
        /// Gets or sets the sample wait time.
        /// </summary>
        /// <value>The sample wait time.</value>
        public int SampleWaitTime
        {
            get
            {
                return sampleWaitFirstHalfTime + sampleWaitSecondHalfTime;
            }
            set
            {
                int oldCycle = CycleWaitTime;
                sampleWaitFirstHalfTime = value / 2;
                if (sampleWaitFirstHalfTime < 1)
                {
                    sampleWaitFirstHalfTime = 1;
                    sampleWaitSecondHalfTime = 1;
                }
                else
                {
                    sampleWaitSecondHalfTime = value - sampleWaitFirstHalfTime;
                }
                CycleWaitTime = oldCycle;
            }
        }
        /// <summary>
        /// Gets or sets the cycle wait time.
        /// </summary>
        /// <value>The cycle wait time.</value>
        public int CycleWaitTime
        {
            get
            {
                return cycleWaitFirstTimePart + cycleWaitSecondTimePart;
            }
            set
            {
                cycleWaitFirstTimePart = sampleWaitFirstHalfTime;
                cycleWaitSecondTimePart = value - cycleWaitFirstTimePart;
                if (cycleWaitSecondTimePart < 1) cycleWaitSecondTimePart = 1;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NTPClient"/> class.
        /// </summary>
        /// <param name="sntpSamples">The SNTP samples.</param>
        /// <param name="sntpHistory">The SNTP history.</param>
        /// <param name="sampleWaitTime">The sample wait time.</param>
        /// <param name="cycleWaitTime">The cycle wait time.</param>
        /// <param name="offset">The offset.</param>
        /// <param name="skew">The skew.</param>
        /// <param name="accuracy">The accuracy.</param>
        public NTPClient(int sntpSamples, int sntpHistory, int sampleWaitTime, int cycleWaitTime, double offset, double skew, int accuracy)
        {
            this.sntpHistory = sntpHistory < 2 ? 2 : sntpHistory;
            this.sntpSamples = sntpSamples < 1 ? 1 : sntpSamples;
            SampleWaitTime = sampleWaitTime;
            CycleWaitTime = cycleWaitTime;
            this.offset = offset;
            this.skew = skew;
            skewMinimum = skew;
            skewMaximum = skew;
            HighResolutionDateTime.Initialize(accuracy);
            baseTime = NTPTime.Now;
            offsetAdjustment = 0;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NTPClient"/> class.
        /// </summary>
        /// <param name="sntpSamples">The SNTP samples.</param>
        /// <param name="sntpHistory">The SNTP history.</param>
        /// <param name="sampleWaitTime">The sample wait time.</param>
        /// <param name="cycleWaitTime">The cycle wait time.</param>
        /// <param name="wiskers">The wiskers.</param>
        public NTPClient(int sntpSamples, int sntpHistory, int sampleWaitTime, int cycleWaitTime):this(sntpSamples, sntpHistory, sampleWaitTime, cycleWaitTime, 0, 0, 10){}
        /// <summary>
        /// Initializes a new instance of the <see cref="NTPClient"/> class.
        /// </summary>
        /// <param name="sampleWaitTime">The sample wait time.</param>
        /// <param name="cycleWaitTime">The cycle wait time.</param>
        /// <param name="wiskers">The wiskers.</param>
        public NTPClient(int sampleWaitTime, int cycleWaitTime):this(15,30,sampleWaitTime,cycleWaitTime){}
        /// <summary>
        /// Initializes a new instance of the <see cref="NTPClient"/> class.
        /// </summary>
        public NTPClient():this(2000, 60000){}

        /// <summary>
        /// Starts this instance.
        /// </summary>
        public void Start()
        {
            lock (this)
            {
                synchronizing = true;
                synchronizerThread = new Thread(new ThreadStart(Synchronizer));
                synchronizerThread.IsBackground = true;
                synchronizerThread.Name = "NTPSynchronizer";
                synchronizerThread.Priority = ThreadPriority.AboveNormal;
                synchronizerThread.Start();
            }
        }

        /// <summary>
        /// Stops this instance.
        /// </summary>
        public void Stop()
        {
            if (synchronizerThread != Thread.CurrentThread)
                lock (this)
                {
                    added.Set();
                    synchronizing = false;
                    if (synchronizerThread != null)
                    {
                        if (synchronizerThread.ThreadState != ThreadState.Stopped)
                        {
                            synchronizerThread.Abort();
                            synchronizerThread.Join();
                            synchronizerThread = null;
                        }
                    }
                }
        }

        /// <summary>
        /// Occurs together with NTP synchronization-progress].
        /// </summary>
        public event NTPClientProgressDelegate Progress;

        private void OnProgress(NTPServerInfo server, NTPClientProgressKind kind)
        {
            if (Progress != null)
                Progress(server ?? (object) this, new NTPClientProgressEventArgs(kind));
        }

        private void Synchronizer()
        {
            try
            {
                int n = 0;
                int m = 0;
                int k = 0;
                long hours;
                bool doneOnce=false;
                double[] samples;
                double[] cycles;
                double samplesAverage = 0;
                double cyclesAverage;
                double cyclesMinimum = 0;
                double cyclesMaximum = 0;	
                NTPTime time1;
                double timeSpan;
                Random rand=new Random();
                
                samples=new double[sntpSamples];
                cycles=new double[sntpHistory];
                if (skew!=0)
                {
                    k = sntpHistory;
                    for (int i = 0; i < k; i++) cycles[i] = skew;
                }
                
                OnProgress(null, NTPClientProgressKind.Synced);
                
                while (synchronizing)
                {
                    if (servers.Count == 0)
                    {
                        added.Reset();
                        added.WaitOne();
                    }
                    NTPServerInfo server = null, first = null;
                    try
                    {
                        server=servers[rand.Next(0,servers.Count)];
                        OnProgress(server, NTPClientProgressKind.Pinging);
                        if (server.Suspensions > 0)
                        {
                            if (!(server.Misses > server.MissesAlarm * server.Delay))
                            {
                                OnProgress(server, NTPClientProgressKind.Blocked);
                                server.Suspensions--;
                                log.InfoFormat("Server {0} was hit, but blocked... {1} hits left.", server.Address, server.Suspensions);
                            }
                            first = server;
                            //Random direction for linear search
                            var index = servers.IndexOf(server) + (int)(hours = (long)rand.Next(0, 2) * -2 + 1);
                            if (index >= servers.Count) index = 0;
                            if (index < 0) index = servers.Count - 1;
                            log.InfoFormat("Searching {0}.", hours > 0 ? "upwards" : "downwards");
                            while ((server = servers[index]).Suspensions > 0 && server != first)
                            {
                                index += (int)hours;
                                if (index >= servers.Count) index = 0;
                                if (index < 0) index = servers.Count - 1;
                            }
                        }
                    }
                    catch (ArgumentOutOfRangeException){}
                    if (server!=first)
                    {
                        try
                        {
                            NTPServerResponse response = GetServerTime(server.Address);
                            OnProgress(server, NTPClientProgressKind.Pinged);
                            server.Misses=0;				
                            samples[n]=response.LocalClockOffset;

                            //Calculate precision workout
                            if (n > 1)
                            {
                                //Calculate FTA median
                                samplesAverage = 0;
                                var local=new List<double>();
                                for (var i = 0; i < n+1; i++) local.Add(samples[i]);
                                local.Sort();
                                samplesAverage = local[local.Count/2];
                            }
                            else if (n > 0) samplesAverage = (samples[0] + samples[1])/2;
                            else samplesAverage = samples[0];

                            log.Info(String.Format("Offset:{0,6:F3}s  Trip:{1,6:F3}s  Avg.:{3,7:F4}s [{4}. {2}]", response.LocalClockOffset, response.RoundTripDelay, server.Address, samplesAverage, n));

                            n++;
                        }
                        catch(Exception ex)
                        {
                            OnProgress(server, NTPClientProgressKind.Failed);
                            log.Error("Failed to retreive time from server "+server.Address, ex);
                            if (server.Misses<server.MissesMaximum)
                            {
                                server.Misses++;
                                log.ErrorFormat("{0} fails left for this server...", server.MissesMaximum - server.Misses);					
                            }
                            else
                            {					
                                server.Suspensions=server.Delay*++server.Misses;
                                log.ErrorFormat("This server is blocked for {0} hits...", server.Suspensions);
                            }
                        }
                    }
                    else
                    {
                        OnProgress(null, NTPClientProgressKind.Offline);
                        log.Error("No more unblocked server available!");
                    }		

                    if (n>=sntpSamples)
                    {
                        //Calculate history average
                        time1 = NTPTime.Now;
                        timeSpan = time1 - baseTime;

                        if (doneOnce)
                        {
                            cycles[m] = skew + samplesAverage/timeSpan;

                            if (k>1)
                            {
                                cyclesMinimum=cycles[0];
                                cyclesMaximum=cycles[0];
                                cyclesAverage=cycles[0];
                                for(int i=1;i<k;i++)
                                {
                                    if (cycles[i]<cyclesMinimum) cyclesMinimum=cycles[i];												
                                    if (cycles[i]>cyclesMaximum) cyclesMaximum=cycles[i];
                                    cyclesAverage+=cycles[i];
                                }
                                cyclesAverage/=k+1;
                            }
                            else if (k>0)
                            {
                                cyclesAverage=(cycles[0]+cycles[1])/2;
                                if (cycles[0] > cycles[1])
                                {
                                    cyclesMinimum = cycles[0];
                                    cyclesMaximum = cycles[1];
                                }
                                else
                                {
                                    cyclesMinimum = cycles[1];
                                    cyclesMaximum = cycles[0];
                                }
                            }
                            else
                            {
                                cyclesAverage = cyclesMinimum = cyclesMaximum = cycles[0];
                            }

                            lock(this)
                            {
                                offset += CycleWaitTime*offsetAdjustment;
                                offsetAdjustment = (samplesAverage + timeSpan * skew)/CycleWaitTime;
                                baseTime = time1;
                                skew=cyclesAverage;
                                skewMinimum = cyclesMinimum;
                                skewMaximum = cyclesMaximum;
                            }

                            log.InfoFormat(
                                "Skew:{0}us/s (min.{2} / max.{1}) Used:{3}us/s  Adjustment:{4}ns/s",
                                (int)(cycles[m] * 1E6),
                                (int)(cyclesMaximum * 1E6),
                                (int)(cyclesMinimum * 1E6),
                                (int)(cyclesAverage * 1E6),
                                (int)(offsetAdjustment * 1E9)
                                );

                            if (++m >= sntpHistory) m = 0;
                            if (k < sntpHistory) k++;
                        }
                        else
                        {
                            lock(this)
                            {
                                offsetAdjustment = 0;
                                offset += samplesAverage + timeSpan * skew;
                                baseTime = time1;
                            }
                            log.Info("Startup SNTP period finished.");
                        }
                        doneOnce=true;
                        n=0;
                        Thread.Sleep(cycleWaitFirstTimePart);
                        OnProgress(null, NTPClientProgressKind.Idle);
                        OnProgress(null, NTPClientProgressKind.Synced);
                        Thread.Sleep(cycleWaitSecondTimePart);
                    }
                    else
                    {
                        Thread.Sleep(sampleWaitFirstHalfTime);
                        OnProgress(null, NTPClientProgressKind.Idle);
                        Thread.Sleep(sampleWaitSecondHalfTime);
                    }
                }
                synchronizerThread = null;
            }
            catch (ThreadAbortException) { }
        }
    }
}
